/*
 * ya.h
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This file is part of libya.
 *
 * libya is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libya is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libya.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __YA_H__ // {
#define __YA_H__
#include <windows.h>
#include <mmreg.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <winamp/wa_ipc.h>
#include <winamp/out.h>
#ifdef __cpluplus
extern "C" {
#endif

///////////////////////////////////////////////////////////////////////////////
#define YA_ABOUT
//#define YA_THREAD_CO_INITIALIZE

///////////////////////////////////////////////////////////////////////////////
// win32 thread-safe queue implementation using native windows API
// http://stackoverflow.com/questions/11706985/win32-thread-safe-queue-implementation-using-native-windows-api

// Mutex vs. Semaphores – Part 1: Semaphores
// https://blog.feabhas.com/2009/09/mutex-vs-semaphores-%E2%80%93-part-1-semaphores/

// Mutex vs. Semaphores – Part 2: The Mutex
// https://blog.feabhas.com/2009/09/mutex-vs-semaphores-%E2%80%93-part-2-the-mutex/

// Mutex vs. Semaphores – Part 3 (final part): Mutual Exclusion Problems
// https://blog.feabhas.com/2009/10/mutex-vs-semaphores-%E2%80%93-part-3-final-part-mutual-exclusion-problems/

// Task Synchronisation
// https://blog.feabhas.com/2009/10/task-synchronisation/

// Task Synchronisation – Part 2: Multiple Tasks and RTOS APIs
// https://blog.feabhas.com/2009/11/task-synchronisation-part-2-multiple-tasks-and-rtos-apis/

///////////////////////////////////////////////////////////////////////////////
#if defined (_MSC_VER) && ! defined (__func__) // {
#define __func__                __FUNCTION__
#endif // }

#define YA_WIDEN2(x)            L ## x
#define YA_WIDEN(x)             YA_WIDEN2(x)

#if ! defined (__WFILE__) // {
#define __WFILE__               YA_WIDEN(__FILE__)
#endif // }

#if ! defined (__wfunc__) // {
#define __wfunc__               YA_WIDEN(__func__)
#endif // }

// http://stackoverflow.com/questions/2192416/how-to-convert-concatenated-strings-to-wide-char-with-the-c-preprocessor
#define YA_STR1(x)              #x
#define YA_STR(x)               YA_STR1(x)

#define YA_WSTR1(x)             L ## #x
#define YA_WSTR(x)              YA_WSTR1(x)

// http://stackoverflow.com/questions/7795514/evaluate-preprocessor-token-before-concatenation
#define YA_CONCAT1(A,B)         A ## B
#define YA_CONCAT(A,B)          YA_CONCAT1(A,B)

///////////////////////////////////////////////////////////////////////////////
#define DUNKNOWN(y) \
  message(1,0,&y,"UNKNOWN: \"%s\" (%d).\n",basename(__FILE__),__LINE__)
#define DERROR(x,y) \
  message(0,x,&y,#x ": \"%s\" (%d).\n",basename(__FILE__),__LINE__)
#define DERRORV(x,y,format,...) \
  message(0,x,&y,#x ": \"%s\" (%d).\n",basename(__FILE__),__LINE__,__VA_ARGS__)

#if defined (YA_DEBUG) // {
  #define DPUTS(level,cs) \
      tputs(&trace,level,cs)
  #define DPRINTF(level,cs,...) \
      tprintf(&trace,level,cs,__VA_ARGS__)
  #define DPUTWS(level,ws) \
      tputws(&trace,level,ws)
  #define DWPRINTF(level,ws,...) \
      twprintf(&trace,level,ws,__VA_ARGS__)
  #define DMESSAGE(m) tprintf(&trace,0,"Error " m ": " \
      "\"%s\" (%d).\n",basename(__FILE__),__LINE__)
  #define DMESSAGEV(m,...) tprintf(&trace,0,"Error " m \
      ": \"%s\" (%d).\n",__VA_ARGS__,basename(__FILE__),__LINE__)
  #define DWMESSAGEV(m,...) twprintf(&trace,0,L"Error " m \
      L": \"%s\" (%d)\n",__VA_ARGS__,wbasename(__WFILE__),__LINE__)
#else // } {
  #define DPUTS(level,cs)
  #define DPRINTF(level,cs,...)
  #define DPUTWS(level,ws)
  #define DWPRINTF(level,ws,...)
#if 0 // {
  #define DMESSAGE(m) message(1,0,NULL,"Error " m ": " \
      "\"%s\" (%d).\n",basename(__FILE__),__LINE__)
  #define DMESSAGEV(m,...) message(1,0,NULL,"Error " m \
      ": \"%s\" (%d).\n",__VA_ARGS__,basename(__FILE__),__LINE__)
#else // {
  #define DMESSAGE(m)
  #define DMESSAGEV(m,...)
#endif // }
  #define DWMESSAGEV(m,...)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YA_GLOBAL_ALLOC) // {
  #define YA_MALLOC(size)       GlobalAlloc(GMEM_FIXED,size)
  #define YA_FREE(p)            GlobalFree(p)
#else // } {
  #define YA_MALLOC(size)       malloc(size)
  #define YA_FREE(p)            free(p)
#endif // }

#define YA_INT_MAX(n)           (~(~0ull<<((n)-1)))

///////////////////////////////////////////////////////////////////////////////
__declspec( dllexport ) Out_Module * winampGetOutModule();
char *basename(char *s);
wchar_t *wbasename(wchar_t *s);
wchar_t *yapath(wchar_t *file, HWND hWnd);
void yapath_free(wchar_t *path);
int message(int force, HRESULT x, HRESULT *y, const char *m, ...);
int crash(const char *format, ...);
#if defined (YA_DEBUG) // {
void ConsoleOpen(void);
void ConsoleClose(void);
#endif // }

void WFXXSetup(WAVEFORMATEXTENSIBLE *pwfxx, int samplerate, int numchannels,
    int bitspersamp, BOOL bSurround, BOOL bFloat);

///////////////////////////////////////////////////////////////////////////////
#define VALUE_INT(pData,offs) \
  (*(int *)(((char *)(pData))+(offs)))
#define VALUE_DOUBLE(pData,offs) \
  (*(double *)(((char *)(pData))+(offs)))
#define VALUE_STRING(pData,offs) \
  ((wchar_t *)(((char *)(pData))+(offs)))

#define PROPERTY_INT(pProperty,pData) \
  VALUE_INT(pData,(pProperty)->offs)
#define PROPERTY_DOUBLE(pProperty,pData) \
  VALUE_DOUBLE(pData,(pProperty)->offs)
#define PROPERTY_STRING(pProperty,pData) \
  VALUE_STRING(pData,(pProperty)->offs)

#define YA_PROPERTY_SIZE        128
#define YA_GROUP_COMMON         L"plugin"
#define YA_PROPERTY_DEBUG       L"debug"
#define YA_PROPERTY_FILE        L"file"
#define YA_PROPERTY_SLEEP       L"sleep"
#define YA_PROPERTY_VIS         L"visualization"
#define YA_PROPERTY_PAGE        L"page"
#if defined (YA_ABOUT) // {
#define YA_PROPERTY_CONFIGX     L"posx"
#define YA_PROPERTY_CONFIGY     L"posy"
#define YA_PROPERTY_ABOUTX      L"aboutx"
#define YA_PROPERTY_ABOUTY      L"abouty"
#else // } {
#define YA_PROPERTY_POSX        L"posx"
#define YA_PROPERTY_POSY        L"posy"
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (YA_DEBUG) // {
#define IDC_IS_DEBUG            -1
#define IDC_IS_FILE             -2
#define IDC_IS_SLEEP            -3
typedef enum _TraceTag TraceTag;
typedef struct _Trace Trace;
#endif // }

typedef struct _PropertyType PropertyType;
typedef struct _PropertyIOConfig PropertyIOConfig;
typedef struct _Property Property;

typedef struct _Result Result;
typedef struct _Store Store;
typedef struct _Request Request;
typedef struct _QueueStrategy QueueStrategy;
typedef struct _Queue Queue;
typedef struct _Ring Ring;

typedef struct _ControlCheckBoxConfig ControlCheckBoxConfig;
typedef struct _ControlRadioButtonConfig ControlRadioButtonConfig;
typedef struct _ControlSliderConfig ControlSliderConfig;
typedef struct _ControlComboBoxList ControlComboBoxList;
typedef struct _ControlComboBoxConfig ControlComboBoxConfig;
typedef struct _ControlSliderCascadeList ControlSliderCascadeList;
typedef struct _ControlSliderCascadeConfig ControlSliderCascadeConfig;
typedef struct _ControlType ControlType;
typedef struct _Control Control;

typedef struct _IPlayer IPlayer;
typedef struct _Player Player;
typedef struct _PlayerStub PlayerStub;

typedef int64_t PlayerProc(Player *pPlayer, Request *pRequest);
typedef void PlayerEventProc(void *data);

#if defined (YA_DEBUG) // {
///////////////////////////////////////////////////////////////////////////////
enum _TraceTag {
  TRACE_CLOSE,
  TRACE_CONSOLE,
  TRACE_FILE
};

struct _Trace {
  int nDebug;
  int bFile;
  int nSleep;
  const char *label;
  const wchar_t *path;
  TraceTag tag;

  FILE *f;
};

extern Trace trace;

int TraceCreate(Trace *pTrace, const char *label, const wchar_t *path);
void TraceDestroy(Trace *pTrace, int nSleep);

int TraceSwitch(Trace *pTrace);
int TraceFixIdcs(Trace *pTrace, const int *pIdcs);
Trace *TraceDefault();

int tputs(Trace *pTrace, int nLevel, const char *s);
int tputws(Trace *pTrace, int nLevel, const wchar_t *ws);
int tvprintf(Trace *pTrace, const char *format, va_list ap);
int tprintf(Trace *pTrace, int nLevel, const char *format, ...);
int twprintf(Trace *pTrace, int nLevel, const wchar_t *format, ...);
#endif // }

///////////////////////////////////////////////////////////////////////////////
struct _PropertyIOConfig {
  const wchar_t *path;
  void *pData;
  const void *pDefault;
  const wchar_t *group;
  const wchar_t *pfx;
};

struct _PropertyType {
  void (*Load)(const Property *pProperty, PropertyIOConfig *c);
  void (*Save)(const Property *pProperty, const PropertyIOConfig *c);
};

struct _Property {
  wchar_t *key;
  const PropertyType *type;
  int offs;
  int size;
};

extern const PropertyType gcIntType;
extern const PropertyType gcDoubleType;
extern const PropertyType gcStringType;

void PropertySaveInt(const wchar_t *group, const wchar_t *key, int n,
    const wchar_t *path);

void PropertiesLoad(const Property *pProperty, PropertyIOConfig *c);
void PropertiesSave(const Property *pProperty, const PropertyIOConfig *c);

///////////////////////////////////////////////////////////////////////////////
struct _Result {
  Result *prPrev;
  HANDLE hEvent;
  int64_t state;
};

int ResultCreate(Result *pResult);
void ResultDestroy(Result *pResult);

///////////////////////////////////////////////////////////////////////////////
enum {
  STORE_SIZE=12
};

struct _Store {
  Result *prTos;
  Result aResult[STORE_SIZE];
  HANDLE hAvailable;
  HANDLE hMutex;
#if defined (YA_DEBUG) // {
  int nAvailable;
  int nMinAvailable;
#endif // }
};

int StoreCreate(Store *pStore);
void StoreDestroy(Store *pStore);

void StorePush(Store *pStore, Result *pResult);
Result *StorePop(Store *pStore);

void StoreLock(Store *pStore);
void StoreUnlock(Store *pStore);

Result *StoreGet(Store *pStore);
void StorePut(Store *pStore, Result *pResult);

///////////////////////////////////////////////////////////////////////////////
struct _Request {
  Result *pResult;
  int exit;
  PlayerProc *pPlayerProc;
  va_list ap;
};

///////////////////////////////////////////////////////////////////////////////
struct _QueueStrategy {
  const char *name;
  HANDLE (*GetSemaphoreDown)(Queue *pQueue);
  HANDLE (*GetSemaphoreUp)(Queue *pQueue);
  BOOL (*IsAlertable)(Queue *pQueue);
  Request **(*GetRequest)(Queue *pQueue);
#if defined (YA_DEBUG) // {
  void (*Inc)(Queue *pQueue);
  void (*Dec)(Queue *pQueue);
#endif // }
};

extern const QueueStrategy cqsWrite;
extern const QueueStrategy cqsRead;

///////////////////////////////////////////////////////////////////////////////
enum {
  QUEUE_SIZE=12
};

struct _Queue {
  Request aRequest[QUEUE_SIZE];
  Request *rp;
  Request *wp;
  Request *mp;
  HANDLE hAvailable;
  HANDLE hWritten;
  HANDLE hMutex;
#if defined (YA_DEBUG) // {
  int nAvailable;
  int nMinAvailable;
#endif // }
};

int QueueCreate(Queue *pQueue);
void QueueDestroy(Queue *pQueue);

void QueueLockMutex(Queue *pQueue);
void QueueUnlockMutex(Queue *pQueue);

Request *QueueLock(Queue *pQueue, const QueueStrategy *pStrategy,
    HANDLE hEvent, void (*OnEvent)(void *data), void *data);
void QueueUnlock(Queue *pQueue, const QueueStrategy *pStrategy,
    Result *pResult);
Request *QueueLockWrite(Queue *pQueue, Result *pResult, int exit,
    PlayerProc *pPlayerProc, va_list ap);
Request *QueueLockRead(Queue *pQueue, HANDLE hEvent,
    void (*OnEvent)(void *data), void *data);
void QueueUnlockRead(Queue *pQueue, Result *pResult);
void QueueWrite(Queue *pQueue, Store *pStore, int exit,
    PlayerProc *pPlayerProc, va_list ap);

#if defined (YA_DEBUG) // {
void QueueIncAvailable(Queue *pQueue);
void QueueDecAvailable(Queue *pQueue);
void QueueNopAvailable(Queue *pQueue);
#endif // }

///////////////////////////////////////////////////////////////////////////////
struct _ControlType {
  const char *magic;
  const char *name;
  void (*Init)(const Control *pControl, HWND hDlg, HWND hWnd);
  void (*Set)(const Control *pControl, HWND hDlg, const void *pData);
  void (*Get)(const Control *pControl, HWND hDlg, void *pData);
  void (*Sync)(const Control *pControl, HWND hDlg);
};

struct _ControlCheckBoxConfig {
  int idc;
  int offset;
  int off;
  int on;
  wchar_t *help;
};

struct _ControlRadioButtonConfig {
  int idc;
  int offset;
  int val;
  wchar_t *help;
};

struct _ControlSliderConfig {
  int idcSlider;
  int offset;
  double min;
  double max;
  LONG lMaxRange;
  int idcStatic;
  const wchar_t *format;
  wchar_t *help;
};

struct _ControlComboBoxList {
  const wchar_t *label;
  int val;
};

struct _ControlComboBoxConfig {
  int idc;
  int offset;
  const ControlComboBoxList *list;
  wchar_t *help;
};

struct _ControlSliderCascadeList {
  int idcSlider;
  int offset;
  int idcStatic;
  const wchar_t *format;
  wchar_t *help;
};

struct _ControlSliderCascadeConfig {
  double min;
  double max;
  LONG lMaxRange;
  const ControlSliderCascadeList *list;
};

struct _Control {
  const ControlType *type;
  const void *config;
};

extern const ControlType gcCheckBoxType;
extern const ControlType gcRadioButtonType;
extern const ControlType gcSliderType;
extern const ControlType gcComboBoxType;
extern const ControlType gcSliderCascadeType;

HWND ControlCreateToolTip(HWND hDlg, HINSTANCE hInstance);
void ControlAddToolTip(HWND hDlg, HWND hWnd, int idc, PWSTR pszText);

///////////////////////////////////////////////////////////////////////////////
extern const Control gcaCommonControls[];
extern const Control gcaDeviceControls[];
extern const Control gcaCoreDeviceControls[];
extern const Control gcaBuffersControls[];
#if defined (YA_DEBUG) // {
extern const Control gcaTraceControls[];
#endif // }

HWND ControlsInit(const Control *pControl, HWND hDlg, HINSTANCE hInstance);
void ControlsSet(const Control *pControl, HWND hDlg, const void *pData);
void ControlsGet(const Control *pControl, HWND hDlg, void *pData);
void ControlsSync(const Control *pControl, HWND hDlg);

///////////////////////////////////////////////////////////////////////////////
struct _IPlayer {
  PlayerProc *RunProc;
  PlayerProc *StopProc;
  void (*EventProc)(void *pPlayer);
  HANDLE (*GetEvent)(Player *pPlayer);
#if defined (YA_DEBUG) // {
  const int *pTraceIdcs;
#endif // }
};

struct _PlayerStub {
  const IPlayer *lpVtbl;
  Player *pPlayer;
  Store store;
  Queue queue;
  HANDLE hThread;
};

PlayerStub *PlayerStubCreate(const IPlayer *lpVtbl, Player *pPlayer);
void PlayerStubDestroy(PlayerStub *pStub);

DWORD WINAPI PlayerStubThread(LPVOID lpParameter);

int64_t PlayerStubSendV(PlayerStub *pStub, int exit, PlayerProc *pPlayerProc,
    va_list ap);
int64_t PlayerStubSend(PlayerStub *pStub, int exit, PlayerProc *pPlayerProc,
    ...);
void PlayerStubPost(PlayerStub *pStub, int exit, PlayerProc *pPlayerProc);

#ifdef __cpluplus
}
#endif
#endif // }
